"""#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#
#                               Pypes                                          #
#                             ---------                                        #
#                                                                              #
#   Author:                                                                    #
#           Michael Sverdlin                                                   #
#   Date:                                                                      #
#           09.08                                                              #
#   Purpose:                                                                   #
#                                                                              # 
#   TODO:                                                                      #
#        FIX score alignment with time alignment (using spaces?)               #
#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#"""
#------------------------------IMPORTS-----------------------------------------#

import os
import sys
import time
import random
random.seed()

import ConfigParser

import pygame
from pygame.locals import *
from pygame.colordict import THECOLORS as COLORS
pygame.init()       #Initialize all of the pygame parts

import data
import menu
import pipes
from scoreboard import *

#------------------------------GLOBALS-----------------------------------------#

DEBUG = True
FULLSCREEN = 0

MAIN_INI_FILE = "game.ini"
INI_HUD_SECTION = "hud"
INI_GAME_SECTION = "game"
INI_GENERAL_SECTION = "general"
INI_NEXTS_SECTION = "nexts"
INI_SCORE_BOARD_SECTION = "scoreboard"

THEME_FILE = "themes.ini"
THEME_GENERAL_SECTION = "general"
THEME_SECTION_PREFIX = "theme"
THEME_ALPHA = 30


class DIRECTIONS:
    """
    Entry and exit direction, relative to the pipe you are ENTETING.
    """
    LEFT = 1
    UP = 2
    RIGHT = 3
    DOWN = 4
    
    OPPOSITES = {LEFT : RIGHT,
                 UP   : DOWN,
                 RIGHT: LEFT,
                 DOWN : UP}
    
    LIST = [1, 2, 3, 4]

class CUSTOM_EVENT:
    """
    User created events.
    """
    ADVANCE_TIMER = USEREVENT + 1

#------------------------------CLASSES-----------------------------------------#

class Theme:
    def __init__(self, game_name, bg_image_path, intro_text, options):
        self.name = game_name
        self.intro_text = intro_text
        self.bg_image_sur = pygame.image.load(bg_image_path)
        self.bg_image_sur.set_alpha(THEME_ALPHA)
        self.options = options
        
class Level:
    def __init__(self, theme, time_left, level = 0):
        self.theme = theme
        self.level = level
        self.time_left = int(time_left)

class Game:
    """
    """
    
    def __init__(self, game_ini_path):
        """
        Initialize the game class.
        """
        try:
            self.levels_played = 0
            self.options = ConfigParser.ConfigParser()
            self.options.read([game_ini_path])
            
            self.main_options = dict(self.options.items(INI_GENERAL_SECTION))
            
            self.load_themes()
            
        except Exception , error:
            error = "Initializing the game class failed:\n %s" % (error)
            raise
    
    def load_themes(self):
        """
        Load all the theme info into a theme list.
        """
        self.themes_ini = ConfigParser.ConfigParser()
        self.themes_ini.read([data.filepath(THEME_FILE)])
        num_of_themes = self.themes_ini.getint(THEME_GENERAL_SECTION, "theme_count")
        
        self.themes = []
        for i in xrange(num_of_themes):
            cur_theme_options = dict(self.themes_ini.items(THEME_SECTION_PREFIX + "_" + str(i)))
            cur_theme = Theme(cur_theme_options["name"], data.filepath(cur_theme_options["bg"]), cur_theme_options["intro_text"], cur_theme_options)
            self.themes.append(cur_theme)
            
    def start(self):
        """
        Open the display.
        """
        if "true" == self.main_options["full_screen"]:
            FULLSCREEN = pygame.FULLSCREEN
        else:
            FULLSCREEN = 0
            
        self.screen = pygame.display.set_mode((int(self.main_options["screen_width"]), int(self.main_options["screen_height"])), FULLSCREEN | pygame.HWSURFACE | pygame.DOUBLEBUF)
        pygame.display.set_caption("Pypes!")
                
        self.start_level()
    
    def start_level(self):
        """
        Start a level.
        """
        
        score_board_options = dict(self.options.items(INI_SCORE_BOARD_SECTION))
        self.score_board = ScoreBoard(ini_list_to_rect(self.main_options["score_board_rect"]), score_board_options)
        
        self.pipe_generator = PipeGenerator(ini_list_to_rect(self.main_options["nexts_rect"]),  dict(self.options.items(INI_NEXTS_SECTION)), data.filepath("pipes.ini"))
        
        self.level = Level(random.choice(self.themes), self.get_time())
        
        self.game_board = GridClass(self.options, self.score_board, self.pipe_generator, self.level)
        
        self.main_game_loop()
    
    def get_time(self):
        """
        Computes time for the current level.
        """
        return (int(self.main_options["max_levels"]) - self.levels_played) * int(self.main_options["time_mul"])
    
    def main_game_loop(self):
        """
        This will feature the main game loop.
        """
        try:
            self.clock = pygame.time.Clock()
            
            if pygame.cursors.arrow != pygame.mouse.get_cursor():
                pygame.mouse.set_cursor(*pygame.cursors.arrow)
            
            pygame.time.set_timer(CUSTOM_EVENT.ADVANCE_TIMER, 1000)
            self.score_board.show(self.level.time_left)
            self.pipe_generator.show_pipes()
            self.should_run = True
            
            while self.should_run:
                
                for event in pygame.event.get():
                    #Check if we need to quit:
                    if QUIT == event.type:
                        self.should_run = False
                    #Quit if alt+f4 was hit:
                    elif ((KEYDOWN == event.type) and (K_F4 == event.key) and (4352 == event.mod)):
                        pygame.event.post(pygame.event.Event(QUIT))
                    
                    #CHEAT CODES for time manipulations:
                    elif (KEYDOWN == event.type):
                        if (K_KP_PLUS == event.key):
                            self.level.time_left += 5
                            pygame.event.post(pygame.event.Event(CUSTOM_EVENT.ADVANCE_TIMER))
                        elif (K_KP_MINUS == event.key):
                            self.level.time_left -= 5
                            pygame.event.post(pygame.event.Event(CUSTOM_EVENT.ADVANCE_TIMER))
                    
                    #Update the score board:
                    elif CUSTOM_EVENT.ADVANCE_TIMER == event.type:
                        #Draw the score board:    
                        self.level.time_left -= 1
                        
                        if 0 >= self.level.time_left:
                            pygame.time.set_timer(CUSTOM_EVENT.ADVANCE_TIMER, 0)
                            self.level.time_left = 0
                            self.game_board.start_flow()
                        
                        self.score_board.show(self.level.time_left)
                    
                    elif (MOUSEBUTTONDOWN == event.type) and (1 == event.button) and (check_point(event.pos, self.game_board.rect)):
                        self.game_board.handle_input(event)
                    
                    
                    
                #Blit the scores:
                self.screen.blit(self.score_board.surface, self.score_board.rect)
                
                #Blit the game board:
                self.game_board.blit()
                self.screen.blit(self.game_board.surface, self.game_board.rect)
                
                self.pipe_generator.show_pipes()
                self.screen.blit(self.pipe_generator.surface, self.pipe_generator.rect)
                
                #Blit the bg of the theme:
                self.screen.blit(self.level.theme.bg_image_sur, Rect(0, 0, 600, 800))
                
                pygame.display.update()
                self.clock.tick(100)
                #print self.clock.get_fps()

        except Exception , error:
            error = "_main_game_loop in the game class failed:\n %s" % (error)
            raise
        
        
class PypeContainer:
    """
    A class representing a single network pipe (TWISTED PAIR!)
    """
    HEIGHT = 50
    WIDTH = 50
    
    def __getattr__(self, attr):
        """
        Some attributes need to go to the tile...
        """
        return getattr(self.tile, attr)
    
    def __init__(self, location_rect, tile):
        self.rect = location_rect
        self.surface = tile.surface
        self.tile = tile
        self.entered = False
        
        #Fill with eandom colors :):
        #self.surface.fill(random.choice(COLORS.values()))
        
    def enter(self, direction):
        print "ti>>", self.tile.e2e, direction
        if self.tile.e2e.has_key(DIRECTIONS.OPPOSITES[direction]):
            return True
        
        return False
        
    
class GridClass:
    """
    The class representing the gameboard.
    """
    
    def __init__(self, all_options, score_board, pipe_generator, level):
        self.level = level
        self.whole_ini = all_options
        self.options = dict(self.whole_ini.items(INI_GAME_SECTION))
        self.pipe_generator = pipe_generator
        
        self.is_flowing = False
        self.is_lost = False
        
        self.rect = ini_list_to_rect(dict(self.whole_ini.items(INI_GENERAL_SECTION))["gameboard_rect"])
        self.surface = pygame.Surface((self.rect.width, self.rect.height))
        self.surface.fill(COLORS["red"])
        
        self.score_board = score_board
        
        self.columns = self.rect.width / PypeContainer.WIDTH
        self.rows = self.rect.height / PypeContainer.HEIGHT
        print "Columns %s, Rows %s" % (self.columns, self.rows)
        
        self.tiles = []
        for i in xrange(self.rows):
            row = []
            for j in xrange(self.columns):
                x_pos = PypeContainer.WIDTH * j
                y_pos = PypeContainer.HEIGHT * i
                
                pos_rect = Rect(x_pos, y_pos, PypeContainer.WIDTH, PypeContainer.HEIGHT)
                row.append(PypeContainer(pos_rect, pipes.BasePipe()))
            self.tiles.append(row)
        
        self.choose_start()
        self.choose_stop()
    
    def choose_start(self):
        """
        Choose a random place to put the start pipe.
        """
        
        start_pipe_options = dict(self.pipe_generator.pipes_options.items("start_pipe"))
        image_name = start_pipe_options["image"]
        
        exit_direction = str(random.randint(1, 4))
        start_pipe_options["image"] = os.path.join(os.path.dirname(image_name),  exit_direction + "_" + os.path.basename(image_name))
        start_pipe = pipes.StartPipe(exit_direction, start_pipe_options)
        
        x_pos = random.randint(1, self.columns - 2)
        y_pos = random.randint(1, self.rows - 2)
        print "Start pos %s, %s" % (x_pos, y_pos)
        
        self.switch_pipe(x_pos, y_pos, start_pipe)
        self.start_pipe_loc = (x_pos, y_pos)
    
    def choose_stop(self):
        """
        Choose a random place to put the start pipe.
        """
        
        pipe_options = dict(self.level.theme.options)
        image_name = pipe_options["image"]
        
        enter_direction = str(random.randint(1, 4))
        pipe_options["image"] = os.path.join(os.path.dirname(image_name),  enter_direction + "_" + os.path.basename(image_name))
        stop_pipe = pipes.StopPipe(int(enter_direction), pipe_options)
        
        while 1:
            x_pos = random.randint(1, self.columns - 2)
            y_pos = random.randint(1, self.rows - 2)
            if self.not_near_start(x_pos, y_pos):
                break
                
        print "Stop pos %s, %s" % (x_pos, y_pos)
        
        self.switch_pipe(x_pos, y_pos, stop_pipe)
        self.stop_pipe_loc = (x_pos, y_pos)
    
    def not_near_start(self, x, y):
        """
        Is the position x,y in the grid isn't on the start place.
        """
        if (x == self.start_pipe_loc[0]):
            return False
        
        if (y == self.start_pipe_loc[1]):
            return False
        
        return True
    
    def switch_pipe(self, x, y, new_pipe):
        """
        Switch the pipe in the location (x,y) with a new pipe.
        """
        self.tiles[y][x] = PypeContainer(self.tiles[y][x].rect, new_pipe)
    
    def blit(self):
        """
        Blit the gameboard on the screen.
        """
        
        for i in xrange(self.rows):
            for j in xrange(self.columns):
                self.surface.blit(self.tiles[i][j].surface, self.tiles[i][j].rect)
    
    def handle_input(self, event):
        """
        Handles a mouse input event.
        """
        grid_x = mouse_x_into_tile_x(event.pos[0])
        grid_y = mouse_y_into_tile_y(event.pos[1])
        
        if not (self.tiles[grid_y][grid_x].can_build_on):
            return
        pipe_to_add = self.pipe_generator.get_next()

        self.switch_pipe(grid_x, grid_y, pipe_to_add)
    
    def start_flow(self):
        """
        Called when the network connection starts. 
        """
        self.is_flowing = True
        
        #Get the start tile:
        start_tile = self.tiles[self.start_pipe_loc[1]][self.start_pipe_loc[0]]
        
        next_tile_pos = self.start_pipe_loc[0], self.start_pipe_loc[1]
        next_tile_direction = start_tile.exit
        
        while 1:
            #Calculate the next tile with the postition and the direction of the current:
            last_tile_pos = next_tile_pos
            next_tile_pos = self.calc_next_tile(next_tile_pos, next_tile_direction)
            
            #If there is no such tile you lost:
            if None == next_tile_pos:
                self.spill(last_tile_pos)
                self.is_lost = True
                self.is_flowing = False
                break
            
            elif self.stop_pipe_loc == next_tile_pos:
                self.won(next_tile_pos)
                self.is_flowing = False
                break
            
            chosen_tile = self.tiles[next_tile_pos[1]][next_tile_pos[0]]
            self.score_board.score += int(chosen_tile.score)
            #Find the exit direction using the last direction:
            next_tile_direction = chosen_tile.e2e[DIRECTIONS.OPPOSITES[next_tile_direction]]
    
    def won(self, last_tile_pos):
        """
        ends the level!
        """
        print "I RULE!!"
        self.tiles[last_tile_pos[1]][last_tile_pos[0]].surface.fill(COLORS["blue"])
    
    def calc_next_tile(self, (x, y), direction):
        """
        Find the next tile the network is going to. Return None for a spill.
        """
        print ">>>", x, y, direction
        
        if 1 == direction:
            next_tile_pos = (x - 1, y)
        elif 2 == direction:
            next_tile_pos = (x, y - 1)
        elif 3 == direction:
            next_tile_pos = (x + 1, y)
        elif 4 == direction:
            next_tile_pos = (x, y + 1)
        else:
            raise ValueError("Direction given (%s) is wrong!" % direction)
        
        if (self.rows < next_tile_pos[0]) or (0 > next_tile_pos[0]) or (self.columns < next_tile_pos[1]) or (0 > next_tile_pos[1]):
            return None
        
        next_tile = self.tiles[next_tile_pos[1]][next_tile_pos[0]]
        
        if (not next_tile.is_enterable) or (next_tile.entered):
            return None
        
        if not next_tile.enter(direction):
            return None
        
        return next_tile_pos
        
        
    def spill(self, last_tile_pos):
        """
        Symbolizes a break in the chain where a spill occurs and
        """
        self.tiles[last_tile_pos[1]][last_tile_pos[0]].surface.fill(COLORS["red"])
        
class PipeGenerator:
    """
    This class contains the next pipes availibale to the player and to show that
    pipe.
    """
    
    def __init__(self, rect, options, pipe_config):
        self.rect = rect
        self.options = options
        
        self.pipes_options = ConfigParser.ConfigParser()
        self.pipes_options.read([pipe_config])
        
        self.surface = pygame.Surface((rect.width, rect.height))
        self.surface.set_colorkey(COLORS["black"])
        
        self.pipes_to_show = 5
        
        self.next_pipes = self.generate_some_pipes()
        
    
    def get_next(self):
        """
        Get the next pipe.
        """
        ret_pipe = self.next_pipes[0]
        self.next_pipes = self.next_pipes[1:]
        
        if self.pipes_to_show >= len(self.next_pipes):
            self.next_pipes += self.generate_some_pipes()
        
        return ret_pipe
    
    def generate_some_pipes(self):
        """
        Generate some random pipes to fill up the world.
        """
        next_pipes = []
        for i in xrange(int(self.options["pipes_per_batch"])):
            type = random.choice([1])
            
            if 1 == type:
                #Straight chosen:
                cur_options = dict(self.pipes_options.items("StraightPipe"))
                cur_start_dir = random.choice(DIRECTIONS.LIST)
                while 1:
                    cur_exit_dir = random.choice(DIRECTIONS.LIST)
                    if cur_exit_dir != cur_start_dir:
                        break
                
                cur_image = os.path.join(os.path.dirname(cur_options["image"]), str(cur_start_dir) + "_" + str(cur_exit_dir) + "_" + os.path.basename(cur_options["image"]))
                if not os.path.isfile(data.filepath(cur_image)):
                    cur_image = os.path.join(os.path.dirname(cur_options["image"]), str(cur_exit_dir) + "_" + str(cur_start_dir) + "_" + os.path.basename(cur_options["image"]))
                
                cur_options["image"] = cur_image
                
                next_pipes.append(pipes.StraightPipe(cur_start_dir, cur_exit_dir, cur_options))
                    
        
        return next_pipes
    
    def show_pipes(self):
        """
        Show the first 5 pipes (next pipe first...).
        """
        
        for i in xrange(self.pipes_to_show):
            self.surface.blit(self.next_pipes[i].surface, Rect(50, 150 + 65 * i, 50, 50))
        
#------------------------------FUNCTIONS---------------------------------------#

def check_point((x, y), rect):
    """
    Checks if the point represented by (x,y) is in the area of the rect.
    """
    if (x > rect.left) and (x < rect.right):
        if (y < rect.bottom) and (y > rect.top):
            return True
        
    return False

def ini_list_to_rect(ini_area_string):
    """
    Tries to create a rect from the ini_area_string given.
    The string should be something like that: "x y w h".
    """
    split_measures = ini_area_string.split(" ", 4)
    split_measures = [int(i) for i in split_measures]
    
    if 4 == len(split_measures):
        return Rect(split_measures[0], split_measures[1], split_measures[2], split_measures[3])
    raise ValueError("Rect area (%s) in wrong format, stop changing ini :)" % ini_area_string)

def mouse_x_into_tile_x(x):
    """
    Turns mouse position x into tile position x.
    """
    tile_x = ((x) / PypeContainer.WIDTH) - (150 / PypeContainer.WIDTH)
    return tile_x

def mouse_y_into_tile_y(y):
    """
    Turns mouse position y into tile position y.
    """
    tile_y = ((y) / PypeContainer.HEIGHT) - (100 / PypeContainer.WIDTH)
    return tile_y

#------------------------------MAIN--------------------------------------------#

def main():
    try:
        #Remove the python CL arg, we don't need it...
        if "python" in sys.argv[0]:
            sys.argv = sys.argv[1:]
            
        #Start the game already!
        game_instance = Game(data.filepath(MAIN_INI_FILE))
        game_instance.start()
        
    except Exception, error:
        print "Program failed, here's a detailed explanation:"
        print error
        if DEBUG:
            print "Taceback stack:"
            import traceback
            print traceback.print_exc()